/*-
 * Copyright (C)2014..2018 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: recobe.h,v 1.44 2018/12/23 19:39:37 babolo Exp $
 */

#ifndef  _BABOLO_RECOBE_H_
# define _BABOLO_RECOBE_H_

# ifndef _BABOLO_BLINflag_H_
#  warning this file needs babolo/BLINflag.h as a prerequisite
#  include <babolo/BLINflag.h>
# endif
# ifndef RECOBE_COMPAT
#  warning RECOBE_COMPAT not defined
# endif

__BEGIN_DECLS

# define RECOBE_NOTABLE    0
# define RECOBE_TEXT       1
# define RECOBE_G711       2
# define RECOBE_ETAT       3
# define RECOBE_DOMAX      4

# define RECOBE_PREFLEN    256
# define RECOBE_MAXKOIS    9

typedef struct recobe_conv {                                /*    */
    u_int64_t domain  : 16                                  /*                            */
            , encosrc : 23                                  /*  src , 0 -    */
            , nonesrc : 1                                   /* Src                  */
            , encodst : 23                                  /*  dst , 0 -    */
            , nonedst : 1                                   /* Dst                  */
    ;
} recobe_conv;

typedef struct recobe_chunk {
    u_int32_t           length;
    u_int32_t           position;
    u_char              chunk[0];
} recobe_chunk;

typedef struct recobe_conf  recobe_conf;
typedef struct recobe_chain recobe_chain;
typedef int               (*recobe_chunker)(recobe_chain *chain, recobe_chunk *chunk);
typedef recobe_chunk     *(*recobe_chainer)(recobe_chain *chain);

struct recobe_chain {
    BLIN_flag           flags;
#  define      RECOBE_USED       0x000001
#  define      RECOBE_ORDER      0x000002
#  define      RECOBE_TRACE      0x000004
#  define      RECOBE_CHAINER    0x000008
#  define      RECOBE_MODREADY   0x000010
#  define      RECOBE_USES       0x00FF00
#  define      RECOBE_USE0       0x000100
#  define      RECOBE_USE1       0x000200
#  define      RECOBE_USE2       0x000400
#  define      RECOBE_USE3       0x000800
#  define      RECOBE_USE4       0x001000
#  define      RECOBE_USE5       0x002000
#  define      RECOBE_USE6       0x004000
#  define      RECOBE_USE7       0x008000
#  define      RECOBE_USE8       0x010000
#  define      RECOBE_USE9       0x020000
    union {
        u_int32_t           inblock;
        struct {
            u_int16_t           inwlen;
            u_int16_t           inwnum;
    };  };
    recobe_conv         cnv;
    recobe_conf        *cf;
    recobe_chain       *next;
    union {
        recobe_chunker      convertor;
        recobe_chainer      puller;
    };
    recobe_chunk       *tail;
    void               *converdata;
    const u_char       *bsubst;
    const u_char       *esubst;
    union {
        u_int32_t           outblock;
        struct {
            u_int16_t           outwlen;
            u_int16_t           outwnum;
};  };  };

typedef struct recobe_c_tnm {                               /*                    */
    const char         *encoid;                             /*                 */
    const char       *(*enconamess)[];                      /*                      */
    const char         *printable;                          /*                      */
} recobe_c_tnm;

typedef struct recobe_c_tdm {                               /*           *
                                                             *                          */
    const u_int16_t     recountad;                          /*                    */
    const char         *domname;                            /*                              */
    const recobe_c_tnm *encolist;                           /*                      */
    int               (*codnumbyid)(const char *encoid);    /*     -1     */
    recobe_conv       (*convbyname)(const char *reconm);    /*    reconm    */
    u_char           *(*reconv)(recobe_conv cnv);           /*    conv         */
} recobe_c_tdm;

typedef struct recobe_collection_t {                        /*    */
    const u_int16_t     domcount;                           /*                      */
    recobe_conv       (*cnvby)(const char *nm, int domain); /*    nm        */
    u_char           *(*reconv)(recobe_conv cnv);           /*    conv         */
    const recobe_c_tdm *domains[];                          /*    ,        *
                                                             *   NULL                          */
} recobe_collection_t;

struct recobe_conf {
    BLIN_flag           flags;
/*             RECOBE_ORDER      0x000002 */
/*             RECOBE_TRACE      0x000004 */
#  define      RECOBE_PRINTABLE  0x000100
#  define      RECOBE_PRINTATLE  0x000200
#  define      RECOBE_FULLCHECK  0x000400
#  define      RECOBE_PRELINERA  0x000800
#  define      RECOBE_REVERTIZE  0x010000
#  define      RECOBE_INTERTIZE  0x020000
#  define      RECOBE_STRICT     0x040000
#  define      RECOBE_ERHEXPRT   0x080000
    const u_char       *bsubst;
    const u_char       *esubst;
};

/*****************************************************************************************************/
/*                        =============== G.711 ===============                                      */
/* recobe_G711Atest   buf  l  1KHz c  fi (0..7)   G.711 A-law       */
/* recobe_G711utest   buf  l  1KHz c  fi (0..7)   G.711 u-law       */
                                                                                                   /**/
extern const int16_t             recobe_deG711A[256];    /*   G.711 A-law   */
extern const int16_t             recobe_deG711u[256];    /*   G.711 u-law   */
u_char    recobe_G711A        __P((int16_t sample));     /*     G.711 A-law     */
u_char    recobe_G711u        __P((int16_t sample));     /*     G.711 u-law     */
int       recobe_G711Atest    __P((int fi, u_char *buf, size_t l));                                /**/
int       recobe_G711utest    __P((int fi, u_char *buf, size_t l));                                /**/
                                                                                                   /**/
/*****************************************************************************************************/

/*****************************************************************************************************/
/* recobe_collection                                                                    */
                                                                                                   /**/
extern const recobe_collection_t     recobe_collection;                                            /**/
int           recobe_chainout   __P((recobe_chain *chain, int fd));                                /**/
recobe_chain *recobe_openfile   __P((recobe_conf *cf, const char *filename, u_int32_t chunksize)); /**/
recobe_chain *recobe_openc      __P((recobe_conf *cf, const char *p, int domain, recobe_chain *chain));
recobe_chain *recobe_openk      __P(( recobe_conf *cf                                              /**/
                                    , const char *p                                                /**/
                                    , recobe_chain *chain                                          /**/
                                    , const char *modificator                                      /**/
                                   ));                                                             /**/
                                                                                                   /**/
/*****************************************************************************************************/

/*****************************************************************************************************/
/*                                                                              */
# ifdef RECOBE_INTERNAL
                                                                                                   /**/
#  define RECOBE_CHUNK(A) do{ifBLIN_QO3 recobe_testtail(__FILE__, __func__, __LINE__, (A));}while(0)
#  define RECOBE_CHAIN(A) do{ifBLIN_QO3 recobe_testchain(__FILE__, __func__, __LINE__, (A));}while(0)
extern const char      recobe_nullprintable[256];                                                  /**/
extern const char      recobe_koi8printable[256];                                                  /**/
extern const char     *recobe_1k[65536];                                                           /**/
void recobe_testchain    __P((const char *file, const char *func, int line, recobe_chain *chain)); /**/
void recobe_testtail     __P((const char *file, const char *func, int line, recobe_chunk *tail));  /**/
void recobe_cnverr       __P(( const char *file                                                    /**/
                             , const char *func                                                    /**/
                             , int line                                                            /**/
                             , const char *mesg                                                    /**/
                             , recobe_conv cnv                                                     /**/
                            ));                                                                    /**/
int  recobe_inword       __P((recobe_chain *chain, recobe_chunk *chunk, u_int64_t *wchr));         /**/
void recobe_outword      __P((recobe_chain *chain, recobe_chunk *chunk, u_int64_t wchr));          /**/
                                                                                                   /**/
# endif /* RECOBE_INTERNAL */
/*****************************************************************************************************/

/*****************************************************************************************************/
/*   octets      .                                        */
/* 0 > len                                                       */
/* recobe_load       octets  |len|            */
/* recobe_loadM     octets    mask                                            */
/* recobe_store    c     octets  |len|                   */
/* recobe_storeM   c     octets   mask                            */
/* recobe_swap        c   m                                           */
/* recobe_count                                                     */
                                                                                                   /**/
extern const u_int8_t            recobe_cnt1[256];                     /*      */
u_int32_t recobe_crc32     __P((const u_char *octets, ssize_t len));   /*  crc32 (IEEE 802.3) */
u_int32_t recobe_crc24     __P((const u_char *octets, ssize_t len));   /*  crc24 (RFC-2440)   */
u_int8_t  recobe_csum1     __P((const u_char *octets, ssize_t len));   /*     */
u_int64_t recobe_load      __P((const u_char *octets, ssize_t len));                               /**/
void      recobe_store     __P((u_char *octets, ssize_t len, u_int64_t c));                        /**/
u_int64_t recobe_loadM     __P((const u_char *octets, const char *mask));                          /**/
int       recobe_storeM    __P((u_char *octets, const char *mask, u_int64_t c));                   /**/
u_int64_t recobe_uload     __P((const u_char *octets, ssize_t *len));  /* Load from UTF-8            */
int       recobe_ustore    __P((u_char *octets, u_int64_t c));         /* Store  UTF-8              */
int       recobe_ulen      __P((u_char octet));                        /* Stored UTF-8 byte size     */
u_int64_t recobe_cload     __P((const u_char *octets, ssize_t *len));  /* Load from CLEN             */
int       recobe_cstore    __P((u_char *octets, u_int64_t c));         /* Store  CLEN               */
int       recobe_clen      __P((u_char octet));                        /* Stored CLEN byte size      */
u_int64_t recobe_swap      __P((u_int64_t c, u_int8_t m));                                         /**/
int       recobe_count     __P((u_int64_t c));                         /* number of '1's             */
int       recobe_wide      __P((u_int64_t c));                         /* c width                    */
                                                                                                   /**/
/*****************************************************************************************************/

__END_DECLS

#endif /* _BABOLO_RECOBE_H_ */
