/*-
 * Copyright (C)2016..2018 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2014..2018 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: wchar.c,v 1.42 2018/12/23 19:39:37 babolo Exp $"

#define BLIN_COMPAT     3
#define RECOBE_COMPAT   4
#define RECOBE_INTERNAL 1

#include <sys/types.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <babolo/BLINflag.h>
#include "recobe.h"

#define WIDE_L 16
#define NO     NULL

static const u_int16_t
/*****************************************************************************************************
 *****************************************************************************************************/
recobe_k1[256] =                                                                                 /****
 *****************************************************************************************************
 *****************************************************************************************************/
{ 0x0000, 0x0001, 0x0002, 0x0003,  0x0004, 0x0005, 0x0006, 0x0007 /* 00 */
, 0x0008, 0x0009, 0x000A, 0x000B,  0x000C, 0x000D, 0x000E, 0x000F
, 0x0010, 0x0011, 0x0012, 0x0013,  0x0014, 0x0015, 0x0016, 0x0017 /* 10 */
, 0x0018, 0x0019, 0x001A, 0x001B,  0x001C, 0x001D, 0x001E, 0x001F

, 0x0020, 0x0021, 0x0022, 0x0023,  0x0024, 0x0025, 0x0026, 0x0027 /* 20 */
, 0x0028, 0x0029, 0x002A, 0x002B,  0x002C, 0x002D, 0x002E, 0x002F
, 0x0030, 0x0031, 0x0032, 0x0033,  0x0034, 0x0035, 0x0036, 0x0037 /* 30 */
, 0x0038, 0x0039, 0x003A, 0x003B,  0x003C, 0x003D, 0x003E, 0x003F

, 0x0040, 0x0041, 0x0042, 0x0043,  0x0044, 0x0045, 0x0046, 0x0047 /* 40 */
, 0x0048, 0x0049, 0x004A, 0x004B,  0x004C, 0x004D, 0x004E, 0x004F
, 0x0050, 0x0051, 0x0052, 0x0053,  0x0054, 0x0055, 0x0056, 0x0057 /* 50 */
, 0x0058, 0x0059, 0x005A, 0x005B,  0x005C, 0x005D, 0x005E, 0x005F

, 0x0060, 0x0061, 0x0062, 0x0063,  0x0064, 0x0065, 0x0066, 0x0067 /* 60 */
, 0x0068, 0x0069, 0x006A, 0x006B,  0x006C, 0x006D, 0x006E, 0x006F
, 0x0070, 0x0071, 0x0072, 0x0073,  0x0074, 0x0075, 0x0076, 0x0077 /* 70 */
, 0x0078, 0x0079, 0x007A, 0x007B,  0x007C, 0x007D, 0x007E, 0x007F

, 0x2500, 0x2502, 0x250C, 0x2510,  0x2514, 0x2518, 0x251C, 0x2524 /* 80 */
, 0x252C, 0x2534, 0x253C, 0x2580,  0x2584, 0x2588, 0x258C, 0x2590
, 0x2591, 0x2592, 0x2593, 0x2320,  0x25A0, 0x2219, 0x221A, 0x2248 /* 90 */
, 0x2264, 0x2265, 0x00A0, 0x2321,  0x00B0, 0x00B2, 0x00B7, 0x00F7

, 0x2550, 0x2551, 0x2552, 0x0451,  0x2553, 0x2554, 0x2555, 0x2556 /* A0 */
, 0x2557, 0x2558, 0x2559, 0x255A,  0x255B, 0x255C, 0x255D, 0x255E
, 0x255F, 0x2560, 0x2561, 0x0401,  0x2562, 0x2563, 0x2564, 0x2565 /* B0 */
, 0x2566, 0x2567, 0x2568, 0x2569,  0x256A, 0x256B, 0x256C, 0x00A9

, 0x044E, 0x0430, 0x0431, 0x0446,  0x0434, 0x0435, 0x0444, 0x0433 /* C0 */
, 0x0445, 0x0438, 0x0439, 0x043A,  0x043B, 0x043C, 0x043D, 0x043E
, 0x043F, 0x044F, 0x0440, 0x0441,  0x0442, 0x0443, 0x0436, 0x0432 /* D0 */
, 0x044C, 0x044B, 0x0437, 0x0448,  0x044D, 0x0449, 0x0447, 0x044A

, 0x042E, 0x0410, 0x0411, 0x0426,  0x0414, 0x0415, 0x0424, 0x0413 /* E0 */
, 0x0425, 0x0418, 0x0419, 0x041A,  0x041B, 0x041C, 0x041D, 0x041E
, 0x041F, 0x042F, 0x0420, 0x0421,  0x0422, 0x0423, 0x0416, 0x0412 /* F0 */
, 0x042C, 0x042B, 0x0417, 0x0428,  0x042D, 0x0429, 0x0427, 0x042A
};

static int
prt(FILE *f, const char *s) {
    int ex = 0;

    for (int i = 0; !!s[i]; ++i) {
        switch(s[i] & 0xFF) {
            case '\n': ex += fprintf(f, "\\n") ; break;
            case '\r': ex += fprintf(f, "\\r") ; break;
            case '\v': ex += fprintf(f, "\\v") ; break;
            case '\f': ex += fprintf(f, "\\f") ; break;
            case '\t': ex += fprintf(f, "\\t") ; break;
            case '\"': ex += fprintf(f, "\\\""); break;
            case '\\': ex += fprintf(f, "\\\\"); break;
            default  : ex += fprintf(f, "%c", s[i]);
    }   }
    return(ex);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
recobe_PrintWideStr(const char *p) {                                                             /****
 *****************************************************************************************************
 *****************************************************************************************************/
    int    ex = EX_OK;
    u_char t[65536];
    int    no = 0;

    switch(p[0]) {
    case 'U':                                                     /*   recobe_ud[] */
        for (int i = 0; i < 256; ++i) {
            if  (i < 0x80) {
                printf("%c {0x80%02X}", i ? ',' : '{', i);
            } else if (i < 0xC0) {
                printf(", {0x40%02X}", i & 0x3F);
            } else if (i == 0xFF) {
                printf(", {0x2000}");
            } else {
                int    mask;
                int    j;

                for (j = 0, mask = 0x80; i & mask; ++j, mask >>= 1) {}
                printf(", {0x8%1X%02X}", j - 1, i & (mask - 1));
            }
            if  (!(~i & 7)) printf(" /* %02X */\n", i & 0x00F8);
        }
        printf("};\n");
        break;
    case 'N':
        no = 1;
        /* FALLTHROUGH */
    case 'K': {                                          /*    recobe_1k[] */
            FILE *nil = fopen("/dev/null", "w");
            int   virtualpos;
            int   realpos;
            int   shift;

            if  (!nil) {
                ifBLIN_QW0("fopen(/dev/null, w)");
                goto out;
            }
            printf("/*-\n"
" * Copyright (C)2018 @BABOLO http://www.babolo.ru/\n"
" * All rights reserved.\n"
" *\n"
" * Redistribution and use in source and binary forms, with or without\n"
" * modification, are permitted provided that the following conditions\n"
" * are met:\n"
" * 1. Redistributions of source code must retain the above copyright\n"
" *    notice, this list of conditions and the following disclaimer.\n"
" * 2. Redistributions in binary form must reproduce the above copyright\n"
" *    notice, this list of conditions and the following disclaimer in the\n"
" *    documentation and/or other materials provided with the distribution.\n"
" *\n"
" * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND\n"
" * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE\n"
" * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE\n"
" * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE\n"
" * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL\n"
" * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS\n"
" * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)\n"
" * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT\n"
" * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY\n"
" * SUCH DAMAGE.\n"
" */\n"
"\n"
"#ident \"@(#) Copyright (C)2018 @BABOLO http://www.babolo.ru/\"\n"
"#ident \"@(#) $Id: wchar.c,v 1.42 2018/12/23 19:39:37 babolo Exp $\"\n"
"\n"
"#include <stdio.h>\n"
"#define NO NULL\n"
"\n\
const char *\n\
/**************************************************************************************************\n\
 **************************************************************************************************/\n\
recobe_1k[65536] =                                                                            /****\n\
 **************************************************************************************************\n\
 **************************************************************************************************/\n\
");
            for (int i = 0; i < 65536; ++i) {
                int next = 0;
                int now = 0;

                if  (((i & 15) < 15) && !!recobe_1k[i + 1]) next = prt(nil, recobe_1k[i + 1]);
                if  (!(i & 15)) {
                    virtualpos = 0;
                    realpos = 0;
                }
                now = (!recobe_1k[i]) ? -1 : prt(nil, recobe_1k[i]);
                shift = realpos - virtualpos + ((0 < now) ? now : 0);
                if  (next > 2) ++shift;
                switch(shift) {
                case 0:
                    if  (!!i && ((!!no && !now) || (0 > now))) {
                        realpos += printf("%c  NO", !i ? '{' : ',');
                        goto caseout;
                    } else {
                        realpos += printf("%c  \"", !i ? '{' : ',');
                    }
                    break;
                case 1:
                    if  (!!i && ((!!no && !now) || (0 > now))) {
                        realpos += printf("%c NO", !i ? '{' : ',');
                        goto caseout;
                    } else {
                        realpos += printf("%c \"", !i ? '{' : ',');
                    }
                    break;
                default:
                    if  (!!i && ((!!no && !now) || (0 > now))) {
                        realpos += printf("%cNO", !i ? '{' : ',');
                        goto caseout;
                    } else {
                        realpos += printf("%c\"", !i ? '{' : ',');
                    }
                    break;
                }
                realpos += prt(stdout, recobe_1k[i]);
                realpos += printf("\"");
            caseout:
                virtualpos += 5;
                if  (!(~i & 0x0000000F)) {
                    if  (!(i & 0xFFFFFFF0)) {
                        printf("\n");
                    } else {
                        shift = realpos - virtualpos;
                        switch(shift) {
                        case 0: printf(" /* %03X */\n", i >> 4); break;
                        case 1: printf( "/* %03X */\n", i >> 4); break;
                        default: printf( "/*%03X */\n", i >> 4); break;
        }   }   }   }   }
        printf("};\n");
        break;
    default:                                            /*    recobe_1k[] */
        for (int i = 0; i < 65536; ++i) t[i] = 0;
        for (int i = 0; i < 256; ++i) t[recobe_k1[i]] = i;
        for (int i = 0; i < 65536; ++i) {
            if  (!!t[i] && !!recobe_1k[i][0] && (!(t[i] == (0xFF & recobe_1k[i][0])) || !!recobe_1k[i][1])) {
                ifBLIN_QX0("*** %d t(%02X) <> recobe(%02X)", i, t[i], recobe_1k[i][0]);
            }
            if  (!i) {
                printf("{ \"\"");
            } else if (!t[i]) {
                printf(", \"\"");
            } else switch(t[i]) {
            case '\n': printf(", \"\\n\"") ; break;
            case '\r': printf(", \"\\r\"") ; break;
            case '\v': printf(", \"\\v\"") ; break;
            case '\f': printf(", \"\\f\"") ; break;
            case '\t': printf(", \"\\t\"") ; break;
            case '\"': printf(", \"\\\"\""); break;
            case '\\': printf(", \"\\\\\""); break;
            default  : printf(", \"%c\"", t[i]);
            }
            if  (!(~i & 15)) printf(" /* %03X */\n", i >> 4);
        }
        printf("};\n");
    }
out:
    return(ex);
}

static const char *
kode_k[] = {"KOI8", NULL};

static const char *
kode_u[] = {"UTF8", NULL};

static const char *
kode_0[] = {"8bit", NULL};

static const char *
kode_1[] = {"UCS2", NULL};

static const char *
kode_2[] = {"UCS3", NULL};

static const char *
kode_3[] = {"UCS4", NULL};

static const char *
kode_4[] = {"UCS5", NULL};

static const char *
kode_5[] = {"UCS6", NULL};

static const char *
kode_6[] = {"UCS7", NULL};

static const char *
kode_7[] = {"UCS8", NULL};

static const char *
kode_8[] = {"UCS9", NULL};

static const char *
kode_9[] = {"UCS10", NULL};

static const char *
kode_T[] = {"text", NULL};

static const char *
kode_B[] = {"base64", NULL};

static const char *
kode_Q[] = {"quoted-printable", NULL};

static const recobe_c_tnm
/*****************************************************************************************************/
encolist[WIDE_L] =                                                                               /****
 *****************************************************************************************************/
{ { NULL, NULL,     NULL}
, { "k" , &kode_k,  NULL}
, { "u" , &kode_u,  NULL}
, { "0" , &kode_0,  NULL}
, { "1" , &kode_1,  NULL}
, { "2" , &kode_2,  NULL}
, { "3" , &kode_3,  NULL}
, { "4" , &kode_4,  NULL}
, { "5" , &kode_5,  NULL}
, { "6" , &kode_6,  NULL}
, { "7" , &kode_7,  NULL}
, { "8" , &kode_8,  NULL}
, { "9" , &kode_9,  NULL}
, { "T" , &kode_T,  NULL}
, { "B" , &kode_B,  NULL}
, { "Q" , &kode_Q,  NULL}
};

static int
/*****************************************************************************************************/
codnumbyid(const char *nm) {                                                                     /****
 *****************************************************************************************************/
    int   r = 0;
    int   i;

    if  (!nm[1]) {
        for (i = 1; i < WIDE_L; ++i) {
            if  (nm[0] == encolist[i].encoid[0]) {
                r = i;
                break;
    }   }   }
    return(r);
}

static recobe_conv
/*****************************************************************************************************/
convbyname(const char *name) {                                                                   /****
 *****************************************************************************************************/
    recobe_conv cnv = {0, 0, 0, 0, 0};
    char s[2] = "0";

    if  (!name || !name[0]) {
        cnv.nonesrc = cnv.nonedst = 1;
    } else {
        s[0] = name[0];
        cnv.encosrc = codnumbyid(s);
        if  (!name[1]) {
            cnv.nonedst = 1;
        } else if (!!name[2]) {
            cnv.encodst = 0;
        } else {
            s[0] = name[1];
            cnv.encodst = codnumbyid(s);
    }   }
    return(cnv);
}

static u_char b64d[] =
{ 0x80, 0x80, 0x80, 0x80 , 0x80, 0x80, 0x80, 0x80 , 0x80, 0x80, 0x80, 0x80 , 0x80, 0x80, 0x80, 0x80
, 0x80, 0x80, 0x80, 0x80 , 0x80, 0x80, 0x80, 0x80 , 0x80, 0x80, 0x80, 0x80 , 0x80, 0x80, 0x80, 0x80
, 0x80, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0x3E , 0xC0, 0xC0, 0xC0, 0x3F
, 0x34, 0x35, 0x36, 0x37 , 0x38, 0x39, 0x3A, 0x3B , 0x3C, 0x3D, 0xC0, 0xC0 , 0xC0, 0x40, 0xC0, 0xC0

, 0xC0, 0x00, 0x01, 0x02 , 0x03, 0x04, 0x05, 0x06 , 0x07, 0x08, 0x09, 0x0A , 0x0B, 0x0C, 0x0D, 0x0E
, 0x0F, 0x10, 0x11, 0x12 , 0x13, 0x14, 0x15, 0x16 , 0x17, 0x18, 0x19, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0
, 0xC0, 0x1A, 0x1B, 0x1C , 0x1D, 0x1E, 0x1F, 0x20 , 0x21, 0x22, 0x23, 0x24 , 0x25, 0x26, 0x27, 0x28
, 0x29, 0x2A, 0x2B, 0x2C , 0x2D, 0x2E, 0x2F, 0x30 , 0x31, 0x32, 0x33, 0xC0 , 0xC0, 0xC0, 0xC0, 0x80

, 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0
, 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0
, 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0
, 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0

, 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0
, 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0
, 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0
, 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0 , 0xC0, 0xC0, 0xC0, 0xC0
};

enum states
{ b0 /* 0       */
, b1 /* 1       */
, b2 /* 2       */
, b3 /* 3       */
, ba /* S=V==   */
, bb /* S==V=   */
, bz /* skip == */
, bs /* skip =  */
, bx
};

static const char statenames[][3] =
{ "b0", "b1", "b2", "b3", "ba", "bb", "bz", "bs", "bx"};

static const char clasnames[][3] =
{ "LS", "LS", "LS", "LS", "LS", "LS", "LS", "LS", "LS", "LS", "LS", "LS", "LS", "LS", "LS", "LS"
, "LS", "LS", "LS", "LS", "LS", "LS", "LS", "LS", "LS", "LS", "Ls", "Ls", "Ls", "Ls", "Ls", "Ls"
, "Ls", "Ls", "Ls", "Ls", "Ls", "Ls", "Ls", "Ls", "Ls", "Ls", "Ls", "Ls", "Ls", "Ls", "Ls", "Ls"
, "Ls", "Ls", "Ls", "Ls", "Ld", "Ld", "Ld", "Ld", "Ld", "Ld", "Ld", "Ld", "Ld", "Ld", "Ly", "Ly"

, "L=", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L."
, "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L."
, "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L."
, "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L."

, "L-", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,"
, "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,"
, "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,"
, "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,"

, "L+", "L#", "L#", "L#", "L#", "L#", "L#", "L#", "L#", "L#", "L#", "L#", "L#", "L#", "L#", "L#"
};

static const char pn[][3] = 
{ "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "- ", "- ", "- ", "- ", "C0", "C1", "C2", "C3"
, "Ca", "Cb", "Cc", "Cd", "- ", "- ", "- ", "Cx"
};

#define C0 0x00080000 /* 1                 */
#define C1 0x00040000 /* 2                 */
#define C2 0x00020000 /* 3                 */
#define C3 0x00010000 /* 4                 */
#define Ca 0x00008000 /* ===                 */
#define Cb 0x00004000 /* ==                  */
#define Cc 0x00002000 /* =                   */
#define Cd 0x00001000 /*                     */
#define Cx 0x00000100 /* error               */
#define C_state  0xFF /*   */

static u_int32_t automa[bx][5] =
/*       0      4      8         C      nl  */
{{    C0|b1, Cx|bx,    b0,    Cx|bx,    bx}/* b0 */
,{    C1|b2, Ca|bx,    b1, Ca|Cx|bx, Ca|bx}/* b1 */
,{    C2|b3, Cb|bs,    b2, Cb|Cx|bx, Cb|bx}/* b2 */
,{ Cd|C3|bx, Cc|bx,    bx, Cc|Cx|bx, Cc|bx}/* b3 */
,{    Cx|bx,    bb,    bx,    Cx|bx,    bx}/* ba */
,{    Cx|bx,    bx,    bx,    Cx|bx,    bx}/* bb */
,{    Cx|bx,    bs, Cx|bx,    Cx|bx, Cx|bx}/* bz */
,{    Cx|bx,    bx, Cx|bx,    Cx|bx, Cx|bx}/* bs */
};

static int
BT(recobe_chain *chain, recobe_chunk *chunk) {
    int           ord = chain->flags & RECOBE_ORDER;
    u_int32_t     control;
    int           state;
    int           ex = 0;
    ssize_t       inpos;
    u_int32_t     ochr;
    u_char       *ichr;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("+ %08X %"BLIN_X, chain->flags, BLIN_I(chain));
    inpos = chain->tail->length - chain->tail->position;
    if  (inpos < 1) {
        ex = -1;
        goto out;
    }
    ichr = &chain->tail->chunk[chain->tail->position];
    blin_stateheader(BLIN_2STA24G, pn);
    for (ochr = 0, state = b0; state < bx; ++ichr, --inpos, chain->tail->position++) {
        control = automa[state][(0 < inpos) ? (b64d[*ichr] >> 6) : 4];
        blin_statebody( BLIN_2STA24G
                      , pn
                      , statenames
                      , clasnames[b64d[*ichr] & 0xFF]
                      , (char*)ichr
                      , inpos
                      , control
                      , state
                      , ochr
                      , b64d[*ichr]
                      );
        if  (control & C0) {
            ochr = b64d[*ichr] << 18;
        }
        if  (control & C1) {
            ochr |= b64d[*ichr] << 12;
        }
        if  (control & C2) {
            ochr |= b64d[*ichr] << 6;
        }
        if  (control & C3) {
            ochr |= b64d[*ichr];
        }
        if  (control & Ca) {
            ifBLIN_QX1("Trunc 1");
            recobe_store(&chunk->chunk[chunk->length], ord ? 1 : -1, ochr >> 16);
            chunk->length += 1;
            ex = -1;
        }
        if  (control & Cb) {
            if  (ochr & 0x0000FFFF) {
                ifBLIN_QX1("Trunc 2");
                recobe_store(&chunk->chunk[chunk->length], ord ? 2 : -2, ochr >> 8);
                chunk->length += 2;
                ex = -1;
            } else {
                recobe_store(&chunk->chunk[chunk->length], ord ? 1 : -1, ochr >> 16);
                chunk->length += 1;
                ex = 1;
        }   }
        if  (control & Cc) {
            if  (ochr & 0x000000FF) {
                ifBLIN_QX1("Trunc 3");
                recobe_store(&chunk->chunk[chunk->length], ord ? 3 : -3, ochr);
                chunk->length += 3;
                ex = -1;
            } else {
                recobe_store(&chunk->chunk[chunk->length], ord ? 2 : -2, ochr >> 8);
                chunk->length += 2;
                ex = 2;
        }   }
        if  (control & Cd) {
            recobe_store(&chunk->chunk[chunk->length], ord ? 3 : -3, ochr);
            chunk->length += 3;
            ex = 3;
        }
        if  (control & Cx) {
            ifBLIN_QX0("B64 syntax");
        }
        state = control & C_state;
    }
out:;
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static const char b64e[64] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static int
TB(recobe_chain *chain, recobe_chunk *chunk) {
    int           ex = 0;
    ssize_t       inpos;
    u_int32_t     ochr;
    u_char       *ichr;
    struct {
        char *in;
        int   max;
        int   cur;
    }            *parm;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("+ %08X %"BLIN_X" %"BLIN_X, chain->flags, BLIN_I(chain), BLIN_I(chain->converdata));
    inpos = chain->tail->length - chain->tail->position;
    if  (inpos < 1) {
        ex = -1;
        goto out;
    }
    if  (!(chain->flags & RECOBE_MODREADY))  {
        chain->flags |= RECOBE_MODREADY;
        if  (!(parm = malloc(sizeof(*parm)))) {
            ifBLIN_QW0("No mem %d", sizeof(*parm));
            ex = -1;
            goto out;
        }
        parm->in = chain->converdata;
        chain->converdata = parm;
        parm->cur = 0;
        parm->max = 0;
        if  (!!parm->in) {
            parm->max = strtoul(parm->in, NULL, 0);
    }   }
    parm = chain->converdata;
    ichr = &chain->tail->chunk[chain->tail->position];
    ochr = *(ichr++) << 16;
    --inpos;
    chain->tail->position++;
    ex = 2;
    if  (0 < inpos) {
        ochr |= *(ichr++) << 8;
        --inpos;
        chain->tail->position++;
        ex = 3;
        if  (0 < inpos) {
            ochr |= *(ichr++);
            --inpos;
            chain->tail->position++;
            ex = 4;
    }   }
    for (int i = 0; i < 4; ++i) {
        if  (i < ex) {
            chunk->chunk[chunk->length++] = b64e[(ochr >> (6 * (3 - i))) & 0x3F];
        } else {
            chunk->chunk[chunk->length++] = '=';
        }
        if  (!!parm->max && (parm->max <= ++parm->cur)) {
            chunk->chunk[chunk->length++] = '\n';
            parm->cur = 0;
    }   }
out:;
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

enum states1
{ aa /* ..      */
, aq /* =       */
, ad /* =X      */
, an /* =nl     */
, as /* #       */
, ap /* #X      */
, ax
};

static const char statenames1[][3] = { "aa", "aq", "ad", "an", "as", "ap", "ax"};

static const char clasnames1[][3] =
{ "Ld", "Ld", "Ld", "Ld", "Ld", "Ld", "Ld", "Ld", "Ld", "Ld", "LS", "LS", "LS", "LS", "LS", "LS"
, "L?", "L?", "L?", "L?", "L?", "L?", "L?", "L?", "L?", "L?", "Ls", "Ls", "Ls", "Ls", "Ls", "Ls"
, "L=", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L.", "L."
, "L#", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,", "L,"
};

static u_char b16[] =
{ 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x30, 0x30, 0x30 , 0x30, 0x30, 0x20, 0x20
, 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20
, 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20
, 0x00, 0x01, 0x02, 0x03 , 0x04, 0x05, 0x06, 0x07 , 0x08, 0x09, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20

, 0x20, 0x0A, 0x0B, 0x0C , 0x0D, 0x0E, 0x0F, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20
, 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20
, 0x20, 0x1A, 0x1B, 0x1C , 0x1D, 0x1E, 0x1F, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20
, 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20

, 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20
, 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20
, 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20
, 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20

, 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20
, 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20
, 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20
, 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20 , 0x20, 0x20, 0x20, 0x20
};

static const char pn1[][3] = 
{ "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "- ", "- ", "- ", "- ", "D0", "D1", "D2", "D3"
, "Ds", "Dd", "- ", "- ", "- ", "- ", "- ", "Dx"
};

#define D0 0x00080000 /* 1                 */
#define D1 0x00040000 /* 2                 */
#define D2 0x00020000 /* 1                 */
#define D3 0x00010000 /* 2                 */
#define Ds 0x00008000 /*                   */
#define Dd 0x00004000 /*                   */
#define Dx 0x00000100 /* error               */

static u_int32_t automa1[ax][6] =
/*       0         1            2            3            q         nl  */
{{    Dd|ax,    Dd|ax,    Dd|   ax,    Dd|   ax,          aq,       ax}/* aa */
,{    D0|ad,    D2|ad,       Dx|ax,          an,    Dd|   ax,    Dx|ax}/* aq */
,{ Ds|D1|ax, Ds|D3|ax, Ds|Dd|Dx|ax, Ds|Dd|Dx|ax, Ds|Dd|Dx|ax, Ds|Dx|ax}/* ad */
,{    Dd|ax,    Dd|ax,    Dd|   ax,          ax,          aq,       ax}/* an */
,{    D0|ap,    D2|ap,          ax,          ax,          ax,       ax}/* as */
,{ Ds|D1|ax, Ds|D3|ax,    Ds|   ax,    Ds|   ax,    Ds|   ax, Ds|   ax}/* ap */
};

static int
QT(recobe_chain *chain, recobe_chunk *chunk) {
    u_int32_t     control;
    int           state;
    int           ex = 0;
    ssize_t       inpos;
    u_int32_t     ochr;
    u_char       *ichr;
    struct {
        char   *in;
        u_char  Q;
    }            *parm;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("+ %08X %"BLIN_X" %"BLIN_X, chain->flags, BLIN_I(chain), BLIN_I(chain->converdata));
    inpos = chain->tail->length - chain->tail->position;
    if  (inpos < 1) {
        ex = -1;
        goto out;
    }
    if  (!(chain->flags & RECOBE_MODREADY))  {
        chain->flags |= RECOBE_MODREADY;
        if  (!(parm = malloc(sizeof(*parm)))) {
            ifBLIN_QW0("No mem %d", sizeof(*parm));
            ex = -1;
            goto out;
        }
        parm->in = chain->converdata;
        chain->converdata = parm;
        parm->Q = 0;
        if  (!!parm->in) {
            parm->Q = *(parm->in);
        } else {
            parm->Q = (chain->flags & RECOBE_USES) >> 8;
    }   }
    parm = chain->converdata;
    ichr = &chain->tail->chunk[chain->tail->position];
    blin_stateheader(BLIN_2STA24G, pn1);
    for (ochr = 0, state = !!parm->Q ? aa : as; state < ax; ++ichr, --inpos, chain->tail->position++) {
        control = automa1[state]
                         [(!!parm->Q && (*ichr == parm->Q)) ? 4 : (0 < inpos) ? (b16[*ichr] >> 4) : 5]
        ;
        blin_statebody( BLIN_2STA24G
                      , pn1
                      , statenames1
                      , clasnames1[b16[*ichr] & 0x3F]
                      , (char*)ichr
                      , inpos
                      , control
                      , state
                      , ochr
                      , b16[*ichr]
                      );
        if  (control & (D0 | D2)) {
            ochr = (b16[*ichr] & 0x0F) << 4;
        }
        if  (control & (D1 | D3)) {
            ochr |= b16[*ichr] & 0x0F;
        }
        if  (control & Ds) {
            chunk->chunk[chunk->length++] = ochr;
        }
        if  (control & Dd) {
            chunk->chunk[chunk->length++] = *ichr;
        }
        if  (control & Dx) {
            ifBLIN_QX1("HEX syntax");
        }
        state = control & C_state;
        ex = 1;
    }
out:;
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

#define Esym   0x00FF
#define Ei 0x00000100 /* copy as is   */
#define Eh 0x00000200 /* copy hex     */
#define Es 0x00000400 /* substitute   */
#define Er 0x00010000 /* ready        */
#define El 0x00020000 /* end of line  */
#define Et 0x00080000 /* test for EOL */

static u_int16_t q16[] =
{ 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0100, 0x0100, 0x0100, 0x0100 , 0x0100, 0x0100, 0x0100, 0x0100
, 0x0100, 0x0100, 0x0100, 0x0100 , 0x0100, 0x0100, 0x0100, 0x0100
, 0x0100, 0x0100, 0x0100, 0x0100 , 0x0100, 0x0100, 0x0100, 0x0100
, 0x0100, 0x0100, 0x0100, 0x0100 , 0x0100, 0x0100, 0x0100, 0x0100

, 0x0100, 0x0100, 0x0100, 0x0100 , 0x0100, 0x0100, 0x0100, 0x0100
, 0x0100, 0x0100, 0x0100, 0x0100 , 0x0100, 0x0100, 0x0100, 0x0100
, 0x0100, 0x0100, 0x0100, 0x0100 , 0x0100, 0x0100, 0x0100, 0x0100
, 0x0100, 0x0100, 0x0100, 0x0100 , 0x0100, 0x0100, 0x0100, 0x0100
, 0x0100, 0x0100, 0x0100, 0x0100 , 0x0100, 0x0100, 0x0100, 0x0100
, 0x0100, 0x0100, 0x0100, 0x0100 , 0x0100, 0x0100, 0x0100, 0x0100
, 0x0100, 0x0100, 0x0100, 0x0100 , 0x0100, 0x0100, 0x0100, 0x0100
, 0x0100, 0x0100, 0x0100, 0x0100 , 0x0100, 0x0100, 0x0100, 0x0200

, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200

, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
, 0x0200, 0x0200, 0x0200, 0x0200 , 0x0200, 0x0200, 0x0200, 0x0200
};

static u_char h16[] = "0123456789ABCDEF";

static int
TQ(recobe_chain *chain, recobe_chunk *chunk) {
    u_int32_t     control;
    int           ex = 0;
    ssize_t       inpos;
    u_int32_t     ochr;
    u_char       *ichr;
    struct {
        char       *in;                          /*                      */
        int         max;                         /*                          */
        int         cur;                         /*                             */
        int         bord;                        /*        */
        u_char      Q;                           /*   HEX                                */
        u_int16_t   tb[256];
    }            *parm;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("+ %08X %"BLIN_X" %"BLIN_X, chain->flags, BLIN_I(chain), BLIN_I(chain->converdata));
    inpos = chain->tail->length - chain->tail->position;
    if  (inpos < 1) {
        ex = -1;
        goto out;
    }
    if  (!(chain->flags & RECOBE_MODREADY))  {
        chain->flags |= RECOBE_MODREADY;
        if  (!(parm = calloc(1, sizeof(*parm)))) {
            ifBLIN_QW0("No mem %d", sizeof(*parm));
            ex = -1;
            goto out;
        }
        parm->in = chain->converdata;
        chain->converdata = parm;
        parm->Q = (chain->flags & RECOBE_USES) >> 8;
        bcopy(q16, parm->tb, sizeof(parm->tb));
        parm->Q = (chain->flags & RECOBE_USES) >> 8;
        parm->tb[parm->Q] = Eh;
        if  (!!parm->in) {
            u_char *p = (u_char*)parm->in;
            u_char  q, r;

            if  (!(b16[*p] & 0x0F)) {
                for (q = *p; q == *(p++); ) {
                    r = *(p++);
                    parm->tb[r] = *(p++) | Es;
            }   }
            if  (!!*p) {
                parm->max = strtoul((char*)p, (char**)&p, 0);
            }
            if  (!!*(p++)) {
                parm->bord = strtoul((char*)p, (char**)&p, 0);
    }   }   }
    parm = chain->converdata;
    ichr = &chain->tail->chunk[chain->tail->position];
    ochr = *(ichr++);
    --inpos;
    chain->tail->position++;
    control = parm->tb[ochr];
    ex = Er | control;
    if  (control & Ei) {
        if  (' ' == ochr) control |= Et;
        chunk->chunk[chunk->length++] = ochr;
        parm->cur++;
    }
    if  (control & Eh) {
        chunk->chunk[chunk->length++] = parm->Q;
        chunk->chunk[chunk->length++] = h16[ochr >> 4];
        chunk->chunk[chunk->length++] = h16[ochr & 0x0F];
        parm->cur += 3;
    }
    if  (control & Es) {
        if  (' ' == (control & Esym)) control |= Et;
        chunk->chunk[chunk->length++] = control & Esym;
        parm->cur++;
    }
    if  (!!parm->max && (parm->max <= parm->cur)) {
        control |= El;
    } else if ((control & Et) && !!parm->bord && (parm->bord <= parm->cur)) {
        u_char c;

        for (int i = parm->cur, j = 0; ; ++i) {
            c = chain->tail->chunk[chain->tail->position + i - parm->cur];
            if  (parm->tb[c] & Eh) j += 2;
            if  ((i + j) >= parm->max) {
                control |= El;
                break;
            }
            if  (' ' == c) break;
    }   }
    if  (control & El) {
        chunk->chunk[chunk->length++] = parm->Q;
        chunk->chunk[chunk->length++] = '\n';
        parm->cur = 0;
    }
out:;
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
un(recobe_chain *chain, recobe_chunk *chunk) {
    int           ex = 0;
    ssize_t       inpos;
    u_int64_t     wchr;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("+ %08X %"BLIN_X, chain->flags, BLIN_I(chain));
    inpos = chain->tail->length - chain->tail->position;
    wchr = recobe_uload(&chain->tail->chunk[chain->tail->position], &inpos);
    if  (0 > inpos) {
        chain->tail->position += ~inpos;
        ex = -1;
    } else {
        recobe_outword(chain, chunk, wchr);
        chain->tail->position += inpos;
    }
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}


static int
nw(recobe_chain *chain, recobe_chunk *chunk) {
    u_int16_t     trail;
    u_int64_t     wchr;
    int           ex;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("+ %08X %"BLIN_X, chain->flags, BLIN_I(chain));
    if  (0 > (ex = recobe_inword(chain, chain->tail, &wchr))) {
        chain->tail->position += -ex;
        goto out;
    }
    trail = 0;
    chain->tail->position += ex;
    if  ((0x010FFFF < wchr) || ((wchr & 0xF800) == 0xD800)) {
        ex = -1;
        goto out;
    }
    if  (0x010000 <= wchr) {
        trail = 0xDC00 + (wchr & 0x3FF);
        wchr = 0xD7C0 + (wchr >> 10);
    }
    recobe_outword(chain, chunk, wchr);
    if  (!!trail) recobe_outword(chain, chunk, trail);
out:;
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
wn(recobe_chain *chain, recobe_chunk *chunk) {
    u_int16_t     trail;
    u_int64_t     wchr;
    int           ex;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("+ %08X %"BLIN_X, chain->flags, BLIN_I(chain));
    if  (0 > (ex = recobe_inword(chain, chain->tail, &wchr))) {
        chain->tail->position += -ex;
        goto out;
    }
    if  (  ((wchr & 0xFC00) == 0xD800)
        && (chain->inwlen * 2 + chain->tail->position > chain->tail->length)
        ) {
        ex = -1;
        chain->tail->position = chain->tail->length;
        goto out;
    }
    chain->tail->position += ex;
    if  ((4 >= chain->outwlen) && ((wchr & 0xFC00) == 0xDC00)) {
        ex = -1;
        goto out;
    }
    if  ((wchr & 0xFC00) == 0xD800) {
        trail = wchr;
        if  ((0 > (ex = recobe_inword(chain, chain->tail, &wchr))) || ((wchr & 0xFC00) != 0xDC00)) {
            ex = -1;
            goto out;
        }
        wchr = (wchr << 10) + trail - 0x035FDC00;
    }
    recobe_outword(chain, chunk, wchr);
out:;
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
nn(recobe_chain *chain, recobe_chunk *chunk) {
    u_int64_t     wchr;
    int           ex;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("+ %08X %"BLIN_X, chain->flags, BLIN_I(chain));
    if  (0 > (ex = recobe_inword(chain, chain->tail, &wchr))) {
        chain->tail->position += -ex;
        goto out;
    }
    chain->tail->position += ex;
    if  ((8 > chain->outwlen) && ((1ULL << (chain->outwlen * 8)) <= wchr)) {
        ex = -1;
        goto out;
    }
    recobe_outword(chain, chunk, wchr);
out:;
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
nk(recobe_chain *chain, recobe_chunk *chunk) {
    u_int64_t     wchr;
    int           ex;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("+ %08X %"BLIN_X, chain->flags, BLIN_I(chain));
    if  (0 > (ex = recobe_inword(chain, chain->tail, &wchr))) {
        chain->tail->position += -ex;
        goto out;
    }
    chain->tail->position += ex;
    if  ((0x00FFFF < wchr) || !recobe_1k[wchr]) {
        ex = -ex;
        goto out;
    };
    for (int k = 0; !!recobe_1k[wchr] && !!recobe_1k[wchr][k]; ++k) {
        recobe_outword(chain, chunk, recobe_1k[wchr][k]);
    }
out:;
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static int
kn(recobe_chain *chain, recobe_chunk *chunk) {
    u_int64_t     wchr;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("+ %08X %"BLIN_X, chain->flags, BLIN_I(chain));
    wchr = recobe_k1[chain->tail->chunk[chain->tail->position++]];
    recobe_outword(chain, chunk, wchr);
    ifBLIN_QX3("- %d", 1);
    return(1);
#   undef blin_internal_flags
}

static int
nu(recobe_chain *chain, recobe_chunk *chunk) {
    u_int64_t     wchr;
    int           outw;
    int           ex;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("+ %08X %"BLIN_X, chain->flags, BLIN_I(chain));
    ex = recobe_inword(chain, chain->tail, &wchr);
    if  (0 > ex) {
        chain->tail->position += -ex;
        goto out;
    }
    outw = recobe_ustore(&chunk->chunk[chunk->length], wchr);
    chain->tail->position += ex;
    if  (0 > outw) {
        ex = -ex;
        goto out;
    }
    chunk->length += outw;
out:;
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static const recobe_chunker
/*****************************************************************************************************/
intercode[WIDE_L][WIDE_L] =                                                                      /****
 *****************************************************************************************************/
/*       k   u   0   1   2   3   4   5   6   7   8   9   T   B   Q  out */
{ { NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO} /* in  */
, { NO, NO, NO, NO, kn, kn, kn, kn, kn, kn, kn, kn, kn, NO, NO, NO} /* k k */
, { NO, NO, NO, NO, NO, un, un, un, un, un, un, un, un, NO, NO, NO} /* u u */
, { NO, NO, nu, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO} /* 0 n */
, { NO, nk, nu, NO, NO, wn, wn, wn, wn, wn, wn, wn, wn, NO, NO, NO} /* 1 w */
, { NO, nk, nu, NO, nw, NO, nn, nn, nn, nn, nn, nn, nn, NO, NO, NO} /* 2 n */
, { NO, nk, nu, NO, nw, nn, NO, nn, nn, nn, nn, nn, nn, NO, NO, NO} /* 3 n */
, { NO, nk, nu, NO, nw, nn, nn, NO, nn, nn, nn, nn, nn, NO, NO, NO} /* 4 n */
, { NO, nk, nu, NO, nw, nn, nn, nn, NO, nn, nn, nn, nn, NO, NO, NO} /* 5 n */
, { NO, nk, nu, NO, nw, nn, nn, nn, nn, NO, nn, nn, nn, NO, NO, NO} /* 6 n */
, { NO, nk, nu, NO, nw, nn, nn, nn, nn, nn, NO, nn, nn, NO, NO, NO} /* 7 n */
, { NO, nk, nu, NO, nw, nn, nn, nn, nn, nn, nn, NO, nn, NO, NO, NO} /* 8 n */
, { NO, nk, nu, NO, nw, nn, nn, nn, nn, nn, nn, nn, NO, NO, NO, NO} /* 9 n */
, { NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, TB, TQ} /* T T */
, { NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, BT, NO, NO} /* B T */
, { NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, NO, QT, NO, NO} /* Q T */
};

static const u_int16_t
/*****************************************************************************************************/
interinwlen[WIDE_L][WIDE_L] =                                                                    /****
 *****************************************************************************************************/
/*       k   u   0   1   2   3   4   5   6   7   8   9   T   B   Q  out */
{ {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* in  */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* k k */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* u u */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 0 n */
, {  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  1,  1,  1} /* 1 w */
, {  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  1,  1,  1} /* 2 n */
, {  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  1,  1,  1} /* 3 n */
, {  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  1,  1,  1} /* 4 n */
, {  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  1,  1,  1} /* 5 n */
, {  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  1,  1,  1} /* 6 n */
, {  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  1,  1,  1} /* 7 n */
, {  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  1,  1,  1} /* 8 n */
, { 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,  1,  1,  1} /* 9 n */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  3,  1} /* T T */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  4,  1,  1} /* B T */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  3,  1,  1} /* Q T */
};

static const u_int16_t
/*****************************************************************************************************/
interoutwlen[WIDE_L][WIDE_L] =                                                                   /****
 *****************************************************************************************************/
/*       k   u   0   1   2   3   4   5   6   7   8   9   T   B   Q  out */
{ {  1,  1,  1,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10,  1,  1,  1} /* in  */
, {  1,  1,  1,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10,  1,  1,  1} /* k k */
, {  1,  1,  1,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10,  1,  1,  1} /* u u */
, {  1,  1,  1,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10,  1,  1,  1} /* 0 n */
, {  1,  1,  1,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10,  1,  1,  1} /* 1 w */
, {  1,  1,  1,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10,  1,  1,  1} /* 2 n */
, {  1,  1,  1,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10,  1,  1,  1} /* 3 n */
, {  1,  1,  1,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10,  1,  1,  1} /* 4 n */
, {  1,  1,  1,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10,  1,  1,  1} /* 5 n */
, {  1,  1,  1,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10,  1,  1,  1} /* 6 n */
, {  1,  1,  1,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10,  1,  1,  1} /* 7 n */
, {  1,  1,  1,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10,  1,  1,  1} /* 8 n */
, {  1,  1,  1,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10,  1,  1,  1} /* 9 n */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  5,  3} /* T T */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  3,  1,  1} /* B T */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  2,  1,  1} /* Q T */
};

static const u_int16_t
/*****************************************************************************************************/
interinwnum[WIDE_L][WIDE_L] =                                                                    /****
 *****************************************************************************************************/
/*       k   u   0   1   2   3   4   5   6   7   8   9   T   B   Q  out */
{ {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* in  */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* k k */
, {  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  1,  1,  1} /* u u */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 0 n */
, {  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  1,  1,  1} /* 1 w */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 2 n */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 3 n */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 4 n */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 5 n */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 6 n */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 7 n */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 8 n */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 9 n */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* T T */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  2,  1,  1} /* B T */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* Q T */
};

static const u_int16_t
/*****************************************************************************************************/
interoutwnum[WIDE_L][WIDE_L] =                                                                   /****
 *****************************************************************************************************/
/*       k   u   0   1   2   3   4   5   6   7   8   9   T   B   Q  out */
{ {  1,  1,  7,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* in  */
, {  1,  1,  7,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* k k */
, {  1,  1,  7,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* u u */
, {  1,  1,  7,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 0 n */
, {  1,  1,  7,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 1 w */
, {  1,  1,  7,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 2 n */
, {  1,  1,  7,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 3 n */
, {  1,  1,  7,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 4 n */
, {  1,  1,  7,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 5 n */
, {  1,  1,  7,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 6 n */
, {  1,  1,  7,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 7 n */
, {  1,  1,  7,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 8 n */
, {  1,  1,  7,  1,  2,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* 9 n */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* T T */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* B T */
, {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1} /* Q T */
};

recobe_chain *
recobe_openk(recobe_conf *cf, const char *p, recobe_chain *next, const char *modificator) {
#   define blin_internal_flags (cf->flags)
    recobe_chain *chain = NULL;
    u_int32_t     code0;

    ifBLIN_QX3("+ w:%s %"BLIN_X, p, BLIN_I(next));
    if  (!(chain = malloc(sizeof(recobe_chain)))) {
        ifBLIN_QW0("malloc(%d)", sizeof(recobe_chain));
        goto out;
    }
    chain->cnv = convbyname(p);
    if  (!!chain->cnv.nonesrc && !!chain->cnv.nonedst) {
        ifBLIN_QX0("No %s recode", p);
        free(chain);
        chain = NULL;
        goto out;
    }
    chain->convertor = intercode[chain->cnv.encosrc][chain->cnv.encodst];
    if  (!chain->convertor) {
        recobe_cnverr(__FILE__, __func__, __LINE__, "No recode", chain->cnv);
        free(chain);
        chain = NULL;
        goto out;
    }
    code0 = codnumbyid("0");
    chain->cf = cf;
    chain->flags = cf->flags & ~(RECOBE_USED | RECOBE_CHAINER | RECOBE_MODREADY);
    chain->tail = NULL;
    chain->next = next;
    chain->converdata = modificator ? strdup(modificator) : NULL;
    chain->bsubst = cf->bsubst;
    chain->esubst = cf->esubst;
    chain->inwlen = interinwlen[chain->cnv.encosrc][chain->cnv.encodst];
    chain->outwlen = interoutwlen[chain->cnv.encosrc][chain->cnv.encodst];
    chain->inwnum = interinwnum[chain->cnv.encosrc][chain->cnv.encodst];
    chain->outwnum = interoutwnum[chain->cnv.encosrc][chain->cnv.encodst];
out:
    RECOBE_CHAIN(chain);
    ifBLIN_QX3("- %"BLIN_X, chain);
    return(chain);
#   undef blin_internal_flags
}

const recobe_c_tdm
/*****************************************************************************************************
 *****************************************************************************************************/
recobe_wide_dm =                                                                                 /****
 *****************************************************************************************************
 *****************************************************************************************************/
{ WIDE_L
, "WIDE"
, encolist
, codnumbyid
, convbyname
, NULL
};
